import time, HTMLParser, zlib
from twisted.python import log
from twisted.protocols import http
from twisted.internet import reactor, protocol
from twisted.application import internet, service
from twisted.web import http, proxy



class AttentionService:
    """Holds service configuration variables.
    
       host: Hostname of clickstream recorder
       port: Port on which to communicate, almost always 80
       path: Path to recorder, starting with '/'
       cookie: Name & value tuple for a cookie, if any is required
    """
    host = 'exampl.com'
    port = 80
    path = '/recorder/'
    cookie = ('AT_USER_ID', 'cookie value')
    
    # file extensions to ignore for plain image & style requests
    ignore = ('gif', 'png', 'jpg', 'js', 'css', 'jpeg', 'ico')



class AttentionClickReporter(http.HTTPClient):
    def __init__(self, body):
        self.body = body

    def connectionMade(self):
        self.sendCommand('POST', AttentionService.path)
        self.sendHeader('Host', '%s:%d' % (AttentionService.host, AttentionService.port))
        self.sendHeader('Content-Length', len(self.body))
        self.sendHeader('Content-Type', 'text/xml')
        self.sendHeader('Cookie', '='.join(AttentionService.cookie))
        self.endHeaders()
        self.transport.write(self.body)
        
    def handleResponse(self, body):
        pass

class AttentionClickReporterFactory(protocol.ClientFactory):

    noisy = False

    def __init__(self, body):
        self.body = body

    def buildProtocol(self, addr):
        return AttentionClickReporter(self.body)

class AttentionClick:
    """A single clickstream entry, based on an HTTP request & response pair.
    """

    noisy = True
    pending = False
    request = {'command': {'command': '', 'rest': '', 'version': ''}, 'headers': {}}
    response = {'status': {'version': '', 'code': '', 'message': ''}, 'headers': {}, 'body': ''}

    def __init__(self):
        pass
        
    def setRequest(self, command, rest, version, headers):
        self.request['command'] = {'command': command, 'rest': rest, 'version': version}
        self.request['headers'] = headers
        
    def setResponseStatus(self, version, code, message):
        self.response['status'] = {'version': version, 'code': code, 'message': message}
        self.pending = True
        
    def setResponseHeader(self, name, value):
        if self.isReportable():
            self.response['headers'][name] = value
        
    def addResponseBody(self, body):
        if self.isReportable() and self.isTextResponse():
            self.response['body'] += body
        
    def reportClick(self):
        """Report the click (maybe; see below).
        """
        if self.isReportable():
            self.pending = False

            reactor.connectTCP(AttentionService.host, AttentionService.port,
                               AttentionClickReporterFactory(self.pingString()))
        
            if self.noisy:
                log.msg('%s %s %s' % (self.request['command']['command'],
                                      self.url(), self.response['status']['code']))
    
    def isReportable(self):
        """Return True if a click should be reported.
        """
        if not self.pending:
            return False

        # look through AttentionService.ignore, a list of blacklisted file extensions
        for extension in ['.'+e.lower() for e in AttentionService.ignore]:
            if self.url()[-len(extension):].lower() == extension:
                self.pending = False

                if self.noisy:
                    log.msg('Discarding ' + self.url())

                return False
        
        return self.pending
    
    def isTextResponse(self):
        """Return True if response was in 2xx or 3xx range, and may be html.
        """
        return self.response['status']['code'][:1] in ('2', '3') \
           and 'text/' in self.response['headers'].get('Content-Type', '')
           
    def url(self):
        """Return a string for the request URL.
        """
        return 'http://%s%s' % (self.request['headers'].get('host', ''),
                                self.request['command']['rest'])
           
    def title(self):
        """Return a string for the response title (if any).
        """
        if self.response['headers'].get('Content-Encoding', '') == 'gzip':
            # if I was smart, I would figure out how to decode gzipped responses
            return ''
        elif self.isTextResponse():
            try:
                tf = TitleFinder()
                tf.feed(self.response['body'])
                return tf.title
            except:
                return ''
        else:
            return ''
            
    def cookies(self):
        """Return a pair of 1/0 flags indicating cookie behavior.
        
           First means a cookie was sent, second means a cookie was received.
        """
        if self.request['headers'].has_key('cookie') or self.request['headers'].has_key('Cookie'):
            sentCookie = 1
        else:
            sentCookie = 0
        
        if self.response['headers'].has_key('cookie') or self.response['headers'].has_key('Cookie'):
            receivedCookie = 1
        else:
            receivedCookie = 0
            
        return (sentCookie, receivedCookie)
           
    def pingString(self):
        """Generate a ping to an AT recorder.
        
        See http://www.attentiontrust.org/extension/spec

            url
            The URL loaded by the browser.
    
            cookie
            1 (true) or 0 (false) indicating if a cookie was sent from
            the browser to the server.
    
            setCookie
            1 (true) or 0 (false) indicating if a cookie was set by the server.
    
            responseCode
            The HTTP response code of the server. Common values are
            200 (success), 301 (redirect), 404 (not found)
    
            method
            The method by which the browser requested the page.
            Typically GET, POST, or PUT.
    
            title
            The title of the retieved page, if any, as specified
            in the page's <title> tag.
    
            date
            The date+time that the browser requested the page, in UTF format.
        """
        
        (cookie, setCookie) = self.cookies()
            
        ping = """<?xml version="1.0"?>
                  <attention xmlns="http://attentiontrust.org/attention/ns#" version="0.11">
                    <httpTransactions>
                      <httpTransaction>
                        <title>%s</title>
                        <url>%s</url>
                        <cookie>%d</cookie>
                        <setCookie>%d</setCookie>
                        <responseCode>%s</responseCode>
                        <method>%s</method>
                        <date>%s</date>
                      </httpTransaction>
                    </httpTransactions>
                  </attention>""" \
                % (self.title(),
                   self.url(),
                   cookie,
                   setCookie,
                   self.response['status']['code'],
                   self.request['command']['command'],
                   time.strftime('%a, %d %b %Y %H:%M:%S GMT', time.gmtime()))
                   
        return ping

class TitleFinder(HTMLParser.HTMLParser):
    """Parser for HTML content, looks for a page title.
    """
    title = ''
    inTitle = False

    def handle_starttag(self, tag, attrs):
        if tag == 'title':
            self.inTitle = True

    def handle_endtag(self, tag):
        if tag == 'title':
            self.inTitle = False

    def handle_data(self, data):
        if self.inTitle:
            self.title = self.title + data

class AttentionProxyClient(proxy.ProxyClient):
    """Based on Twisted ProxyClient, with an AttentionClick instance variable.
    """
    click = None

    def __init__(self, command, rest, version, headers, data, father):
        """Populate self.click with request info.
        """
        # munge the Accept-Encoding header to ensure plain-text responses?
        # see: http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.3
        # headers['accept-encoding'] = 'identity'
        proxy.ProxyClient.__init__(self, command, rest, version, headers, data, father)
        self.click = AttentionClick()
        self.click.setRequest(command, rest, version, headers)

    def handleStatus(self, version, code, message):
        """Add response status to self.click.
        """
        proxy.ProxyClient.handleStatus(self, version, code, message)
        self.click.setResponseStatus(version, code, message)

    def handleHeader(self, key, value):
        """Add response header to self.click.
        """
        proxy.ProxyClient.handleHeader(self, key, value)
        self.click.setResponseHeader(key, value)

    def handleResponsePart(self, buffer):
        """HTTP response body.
        """
        proxy.ProxyClient.handleResponsePart(self, buffer)
        self.click.addResponseBody(buffer)

    def handleResponseEnd(self):
        """Clean up.
        """
        proxy.ProxyClient.handleResponseEnd(self)
        self.click.reportClick()

class AttentionProxyClientFactory(proxy.ProxyClientFactory):
    """Based on Twisted ProxyClientFactory, with a different client class.
    """
    noisy = False

    def buildProtocol(self, addr):
        """Returning an AttentionProxyClient instead of ProxyClient.
        """
        return AttentionProxyClient(self.command, self.rest, self.version,
                                    self.headers, self.data, self.father)

class AttentionProxyRequest(proxy.ProxyRequest):
    """Based on Twisted ProxyRequest, with a different protocol client factory.
    """
    protocols = {'http': AttentionProxyClientFactory}

class AttentionProxy(proxy.Proxy):
    """Based on Twisted Proxy, with a different requestFactory.
    """
    requestFactory = AttentionProxyRequest



application = service.Application('proxy')

f = http.HTTPFactory()
f.protocol = AttentionProxy
proxyService = internet.TCPServer(8080, f)
proxyService.setServiceParent(application)
